{

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
    *                                                                 *
    *   Copyrightcillas 1.0.3                                         *
    *   Copyright  2004  Florian Oldach <oldach@hartwork.org>        *
    *                     Sebastian Pipping <pipping@hartwork.org>    *
    *                                                                 *
    *   --> http://www.hartwork.org                                   *
    *                                                                 *
    *                                                                 *
    *   This source code is released under LGPL.                      *
    *   See LGPL.txt for details.                        2004-09-01   *
    *                                                                 *
    \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

}

UNIT crc_scor;

{ ----------------------------------------------------------------------------------------------------- INTERFACE ----- }
INTERFACE

  USES
    crt,
    crc_std;

  CONST
    NUMBER_OF_SCORE_ENTRIES = 5;

  TYPE
    tScoreEntry = RECORD
      sKey    : tName;
      sName   : tName;
      iKey    : LONGINT;
      iPoints : LONGINT;

    END;
    tScores = ARRAY[ 1..NUMBER_OF_SCORE_ENTRIES ] OF tScoreEntry;

  VAR
    g_Scores     : tScores;
    g_bNewScores : BOOLEAN;

  FUNCTION  read_scores : BOOLEAN;
  PROCEDURE save_scores;
  PROCEDURE update_scores( MyName : tName; MyPoints : WORD );
  FUNCTION  will_join( MyPoints : WORD ) : BOOLEAN;


{ ------------------------------------------------------------------------------------------------ IMPLEMENTATION ----- }
IMPLEMENTATION

  VAR
    fScores: FILE OF tScoreEntry;


{ --------------------------------------------------------------------------------------------------- null_scores ----- }
  PROCEDURE null_scores; { Reset all scores to zero }
    VAR
      i : BYTE;

    BEGIN
      FOR i := 1 TO NUMBER_OF_SCORE_ENTRIES DO
        BEGIN
          g_Scores[ i ].sName    := '<empty>';
          g_Scores[ i ].iPoints  := 0;

        END;

    END;


{ -------------------------------------------------------------------------------------------------- encode_entry ----- }
  PROCEDURE encode_entry( VAR MyEntry : tScoreEntry ); { Encrypt a highscore entry }
    VAR
      i : BYTE;

    BEGIN
      {
      Encryption:
      (1) A random key is generated
      (2) Data is XORed with that key
      (3) Key is also written to file
      }
      
      WITH MyEntry DO
        BEGIN
          iKey := ( RANDOM( 65535 ) + 1 ) OR ( ( RANDOM( 65535 ) + 1 ) SHL 16 );
          sKey := '';

          iPoints := iPoints XOR iKey;
          FOR i := 1 TO MAX_NAME_LENGTH DO
            BEGIN
              sKey       := sKey + CHR( RANDOM( 255 ) + 1 );
              sName[ i ] := CHR( ORD( sName[ i ] ) XOR ORD( sKey[ i ] ) );

            END;

        END;

    END;


{ -------------------------------------------------------------------------------------------------- decode_entry ----- }
  PROCEDURE decode_entry( VAR MyEntry : tScoreEntry ); { Decrypt a highscore entry }
    VAR
      i     : BYTE;

    BEGIN
      {
      Decryption:
      (1) Data is XORed with the key
      }

      WITH MyEntry DO
        BEGIN
          iPoints := iPoints XOR iKey;
          FOR i := 1 TO MAX_NAME_LENGTH DO
            sName[ i ] := CHR( ORD( sName[ i ] ) XOR ORD( sKey[ i ] ) );

        END;

    END;


{ --------------------------------------------------------------------------------------------------- read_scores ----- }
  FUNCTION read_scores : BOOLEAN; { Read scores from file }
    VAR
      i : BYTE;

    BEGIN
      {$I-}
      RESET( fScores );
      {$I+}

      IF ( IORESULT <> 0 ) THEN
        BEGIN
          null_scores;
          read_scores := FALSE;

        END
      ELSE
        BEGIN
          IF ( FILESIZE( fScores ) >= NUMBER_OF_SCORE_ENTRIES ) THEN
            BEGIN
              SEEK( fScores, 0 );
              FOR i := 1 TO NUMBER_OF_SCORE_ENTRIES DO
                BEGIN
                  READ( fScores, g_Scores[ i ] );
                  decode_entry( g_Scores[ i ] );

                END;

              read_scores := TRUE;

            END
          ELSE
            BEGIN
              null_scores;
              read_scores := FALSE;

            END;

          CLOSE( fScores );

        END;

    END;


{ --------------------------------------------------------------------------------------------------- save_scores ----- }
  PROCEDURE save_scores; { Write scores to file }
    VAR
      i : BYTE;

    BEGIN
      {$I-}
      REWRITE( fScores );
      {$I+}

      IF ( IORESULT = 0 ) THEN
        BEGIN
          FOR i := 1 TO NUMBER_OF_SCORE_ENTRIES DO
            BEGIN
              encode_entry( g_Scores[ i ] );
              WRITE( fScores, g_Scores[ i ] );

            END;

          CLOSE( fScores );

        END;

    END;


{ ------------------------------------------------------------------------------------------------- update_scores ----- }
  PROCEDURE update_scores( MyName : tName; MyPoints : WORD ); { Insert new score }
    VAR
      i, j : BYTE;

    BEGIN
      i := 1;
      WHILE ( i <= NUMBER_OF_SCORE_ENTRIES ) DO
        BEGIN
          IF ( MyPoints <= g_Scores[ i ].iPoints ) THEN
            INC( i )

          ELSE
            BEGIN
              { MyPoints > g_Scores[ i ].iPoints }
              FOR j := NUMBER_OF_SCORE_ENTRIES DOWNTO ( i + 1 ) DO
                g_Scores[ j ] := g_Scores[ j - 1 ];

              g_Scores[ i ].iPoints  := MyPoints;
              g_Scores[ i ].sName    := MyName;

              { Break out of while }
              i := NUMBER_OF_SCORE_ENTRIES + 1;

            END; { ELSE }

        END; { WHILE }

      g_bNewScores := TRUE;

    END;


{ ----------------------------------------------------------------------------------------------------- will_join ----- }
  FUNCTION will_join( MyPoints : WORD ) : BOOLEAN;
    BEGIN
      IF ( MyPoints > g_Scores[ NUMBER_OF_SCORE_ENTRIES ].iPoints ) THEN
        will_join := TRUE

      ELSE
        will_join := FALSE;

    END;


{ ---------------------------------------------------------------------------------------------------------- INIT ----- }
BEGIN
  RANDOMIZE;
  ASSIGN( fScores, FILE_SCORES );
  g_bNewScores := FALSE;

END.