{

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *\
    *                                                                 *
    *   Copyrightcillas 1.0.3                                         *
    *   Copyright  2004  Florian Oldach <oldach@hartwork.org>        *
    *                     Sebastian Pipping <pipping@hartwork.org>    *
    *                                                                 *
    *   --> http://www.hartwork.org                                   *
    *                                                                 *
    *                                                                 *
    *   This source code is released under LGPL.                      *
    *   See LGPL.txt for details.                        2004-09-01   *
    *                                                                 *
    \* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

}

UNIT crc_worm;

{ ----------------------------------------------------------------------------------------------------- INTERFACE ----- }
INTERFACE

  USES
    crt,
    crc_std,
    crc_set;

  TYPE
    pWormSlice = ^tWormSlice;
    tWormSlice = RECORD
      x     : BYTE;
      y     : BYTE;
      pPrev : pWormSlice;
      pNext : pWormSlice;

    END;

    tWorm = OBJECT
      iPoints      : WORD;

      iDir         : BYTE;
      iGrowPoints  : WORD;
      iGrowBody    : WORD;
      bCrash       : BOOLEAN;
      bCookie      : BOOLEAN;
      bDirChanged  : BOOLEAN;

      pHead        : pWormSlice;
      pFoot        : pWormSlice;

      FUNCTION  create( MyRoute : tSnakeBirth; VAR MyMap : tMap ) : BOOLEAN;
      PROCEDURE kill;
      FUNCTION  move_head( VAR MyMap : tMap ) : BOOLEAN; { FALSE -> CRASH }
      FUNCTION  move_foot( VAR MyMap : tMap; bIncScore : BOOLEAN ) : BOOLEAN;
      PROCEDURE paint_head;
      PROCEDURE paint_body( MySlice : pWormSlice );
      PROCEDURE paint_foot;

    END;


{ ------------------------------------------------------------------------------------------------ IMPLEMENTATION ----- }
IMPLEMENTATION

{ --------------------------------------------------------------------------------------------------- move_on_map ----- }
  PROCEDURE move_on_map( VAR x, y : BYTE; iDir : BYTE );
    BEGIN
      CASE iDir OF
        DIR_UP:    DEC( y );
        DIR_RIGHT: INC( x );
        DIR_DOWN:  INC( y );
        DIR_LEFT:  DEC( x );

      END;

      { Enable out-of-screen }
      CASE x OF
        0:              x := MAP_WIDTH;
        MAP_WIDTH + 1:  x := 1;

      END;

      CASE y OF
        0:              y := MAP_HEIGHT;
        MAP_HEIGHT + 1: y := 1;

      END;

  END;


{ -------------------------------------------------------------------------------------------------------- occupy ----- }
  FUNCTION occupy( VAR MyMap : tMap; x, y : BYTE ) : CHAR;
    VAR
      chBefore : CHAR;

    BEGIN
      chBefore := MyMap[ x, y ];
      MyMap[ x, y ] := ASCII_SNAKE;
      occupy := chBefore;

    END;


{ -------------------------------------------------------------------------------------------------- append_slice ----- }
  PROCEDURE append_slice( VAR pCur, pOld : pWormSlice; x, y : BYTE );
    BEGIN
      NEW( pCur );
      pCur^.x := x;
      pCur^.y := y;
      pCur^.pPrev := pOld;
      pCur^.pNext := NIL;

      pOld^.pNext := pCur;

      pOld := pCur;

    END;


{ -------------------------------------------------------------------------------------------------- tWorm.create ----- }
  FUNCTION tWorm.create( MyRoute : tSnakeBirth; VAR MyMap : tMap ) : BOOLEAN;
    VAR
      i, j : BYTE;
{      a, b : tBytePoint;}
      x, y : BYTE;

      pCur, pOld : pWormSlice;

    BEGIN
      iGrowPoints  := 0;
      iGrowBody    := 0;
      bCrash       := FALSE;
      bCookie      := FALSE;
      bDirChanged  := FALSE;

      IF NOT between( MyRoute.iCorners, 1, MAX_CORNERS ) THEN
        BEGIN
          create := FALSE;
          EXIT;

        END;

      x := MyRoute.Start.x;
      y := MyRoute.Start.y;

      { Create head }
      IF ( occupy( MyMap, x, y ) = ASCII_SPACE ) THEN
        BEGIN
          NEW( pHead );
          pHead^.x := x;
          pHead^.y := y;
          pHead^.pPrev := NIL;
          pHead^.pNext := NIL;

          pOld := pHead;

        END
      ELSE
        BEGIN
          pHead := NIL;
          create := FALSE;
          EXIT;

        END;

      { Create body from path }
      FOR i := 1 TO MyRoute.iCorners DO
        BEGIN
          FOR j := 1 TO MyRoute.Path[ i ].iDist DO
            BEGIN
              move_on_map( x, y, MyRoute.Path[ i ].iDir );
              IF ( occupy( MyMap, x, y ) = ASCII_SPACE ) THEN
                append_slice( pCur, pOld, x, y )

              ELSE
                BEGIN
                  create := FALSE;
                  kill;
                  EXIT;

                END;

            END; { FOR }

        END; { FOR }

      { Create foot }
      pFoot := pCur;

      { Dir of foot -> append pseudo-foot }
      IF ( ABS( MyRoute.iDirFoot - MyRoute.Path[ MyRoute.iCorners ].iDir ) = 2 ) THEN
        MyRoute.iDirFoot := MyRoute.Path[ MyRoute.iCorners ].iDir;
      move_on_map( x, y, MyRoute.iDirFoot );
      append_slice( pCur, pOld, x, y );

      { Dir of head }
      IF ( MyRoute.iDirHead = MyRoute.Path[ 1 ].iDir ) THEN
        iDir := ( MyRoute.Path[ 1 ].iDir + 2 ) MOD 4
      ELSE
        iDir := MyRoute.iDirHead;

    END;


{ ---------------------------------------------------------------------------------------------------- tWorm.kill ----- }
  PROCEDURE tWorm.kill;
    BEGIN
      WHILE ( pHead <> NIL ) DO
        BEGIN
          pFoot := pHead^.pNext;
          DISPOSE( pHead );
          pHead := pFoot;

        END;

    END;


{ ----------------------------------------------------------------------------------------------- tWorm.move_head ----- }
  FUNCTION tWorm.move_head( VAR MyMap : tMap ) : BOOLEAN;
    VAR
      pTemp : pWormSlice;

    BEGIN
      NEW( pTemp );
      pTemp^.pPrev := NIL;
      pTemp^.pNext := pHead;
      pHead^.pPrev := pTemp;
      pHead := pTemp;

      pHead^.x := ( pHead^.pNext )^.x;
      pHead^.y := ( pHead^.pNext )^.y;
      move_on_map( pHead^.x, pHead^.y, iDir );

      { Check new head and occupy }
      CASE occupy( MyMap, pHead^.x, pHead^.y ) OF
        ASCII_COOKIE:
          BEGIN
            bCookie := TRUE;
            move_head := TRUE;

          END;

        ASCII_SPACE:
          move_head := TRUE;

      ELSE
        bCrash := TRUE;
        move_head := FALSE;

      END;

    END;


{ ----------------------------------------------------------------------------------------------- tWorm.move_foot ----- }
  FUNCTION tWorm.move_foot( VAR MyMap : tMap; bIncScore : BOOLEAN ) : BOOLEAN;
    BEGIN
      IF ( iGrowBody > 0 ) THEN
        BEGIN
          DEC( iGrowBody );
          move_foot := FALSE;

        END
      ELSE
        BEGIN
          { Refresh map }
          MyMap[ pFoot^.x, pFoot^.y ] := ASCII_SPACE;

          { Cut off pseudo-foot }
          DISPOSE( pFoot^.pNext );
          pFoot^.pNext := NIL;
          pFoot := pFoot^.pPrev;
          move_foot := TRUE;

        END;
        
      IF ( iGrowPoints > 0 ) THEN
        BEGIN
          DEC( iGrowPoints );
          IF bIncScore THEN
            iPoints := iPoints + g_Settings.iSpeed;
          
        END;

    END;


{ ---------------------------------------------------------------------------------------------- tWorm.paint_head ----- }
  PROCEDURE tWorm.paint_head;
    CONST
      LINK_UP     = 1;
      LINK_RIGHT  = 2;
      LINK_DOWN   = 4;
      LINK_LEFT   = 8;

    VAR
      ax, ay, bx, by, ox, oy : BYTE; { cx, cy, }
      iPart : BYTE;

    BEGIN
      GOTOXY( pHead^.x, pHead^.y );
      IF bCrash THEN
        WRITE( 'X' )

      ELSE
        BEGIN
          iPart := 0;

          { Head }
          ax := pHead^.x;
          ay := pHead^.y;

          { Sucessor }
          bx := ( pHead^.pNext )^.x;
          by := ( pHead^.pNext )^.y;

          ox := 2 - ax; { ox := 2 + MAP_WIDTH - ax; }
          oy := 2 - ay; { oy := 2 + MAP_HEIGHT - ay; }

          ax := ( ax + ox ) MOD MAP_WIDTH;
          ay := ( ay + oy ) MOD MAP_HEIGHT;
          bx := ( bx + ox ) MOD MAP_WIDTH;
          by := ( by + oy ) MOD MAP_HEIGHT;

          CASE ( ax - bx + 1) OF

            0:
              BEGIN
                WRITE( ASCII_ARROW_LEFT );
                iPart := iPart OR LINK_LEFT;

              END;

            1:
              BEGIN
                IF ( ay - by < 0 ) THEN
                  BEGIN
                    WRITE( ASCII_ARROW_UP );
                    iPart := iPart OR LINK_UP;

                  END
                ELSE
                  BEGIN
                    WRITE( ASCII_ARROW_DOWN );
                    iPart := iPart OR LINK_DOWN;

                  END;

              END;

            2:
              BEGIN
                WRITE( ASCII_ARROW_RIGHT );
                iPart := iPart OR LINK_RIGHT;

              END;

          END; { CASE }

        END; { ELSE }

    END;


{ ---------------------------------------------------------------------------------------------- tWorm.paint_body ----- }
  PROCEDURE tWorm.paint_body( MySlice : pWormSlice );
    CONST
      LINK_UP     = 1;
      LINK_RIGHT  = 2;
      LINK_DOWN   = 4;
      LINK_LEFT   = 8;

    VAR
      ax, ay, bx, by, cx, cy, ox, oy : BYTE;
      iPart : BYTE;

    BEGIN
      iPart := 0;

      { Predecessor }
      ax := ( MySlice^.pPrev )^.x;
      ay := ( MySlice^.pPrev )^.y;

      { Current slice }
      bx := MySlice^.x;
      by := MySlice^.y;

      ox := 2 - ax; { ox := 2 + MAP_WIDTH - ax; }
      oy := 2 - ay; { oy := 2 + MAP_HEIGHT - ay; }

      ax := ( ax + ox ) MOD MAP_WIDTH;
      ay := ( ay + oy ) MOD MAP_HEIGHT;
      bx := ( bx + ox ) MOD MAP_WIDTH;
      by := ( by + oy ) MOD MAP_HEIGHT;

      CASE ( ax - bx + 1) OF

        0:
          iPart := iPart OR LINK_LEFT;

        1:
          IF ( ay - by < 0 ) THEN
            iPart := iPart OR LINK_UP
          ELSE
            iPart := iPart OR LINK_DOWN;

        2:
          iPart := iPart OR LINK_RIGHT;

      END; { CASE }

      { Successor }
      cx := ( MySlice^.pNext )^.x;
      cy := ( MySlice^.pNext )^.y;
      cx := ( cx + ox ) MOD MAP_WIDTH;
      cy := ( cy + oy ) MOD MAP_HEIGHT;

      GOTOXY( MySlice^.x, MySlice^.y );
      CASE ( cx - bx + 1) OF

        0:
          iPart := iPart OR LINK_LEFT;

        1:
          IF ( cy - by < 0 ) THEN
            iPart := iPart OR LINK_UP
          ELSE
            iPart := iPart OR LINK_DOWN;

        2:
          iPart := iPart OR LINK_RIGHT;

      END; { CASE }

      IF g_Settings.bSolidSnakes THEN
        BEGIN
          CASE iPart OF
            LINK_UP OR LINK_RIGHT:
              WRITE( ASCII_CORNER_NE );

            LINK_RIGHT OR LINK_DOWN:
              WRITE( ASCII_CORNER_ES );

            LINK_DOWN OR LINK_LEFT:
              WRITE( ASCII_CORNER_SW );

            LINK_LEFT OR LINK_UP:
              WRITE( ASCII_CORNER_WN );

            LINK_LEFT OR LINK_RIGHT:
              WRITE( ASCII_LINE_HORZ );

            LINK_UP OR LINK_DOWN:
              WRITE( ASCII_LINE_VERT );

          END; { CASE }

        END
      ELSE
        WRITE( ASCII_DOT );

    END;


{ ---------------------------------------------------------------------------------------------- tWorm.paint_foot ----- }
  PROCEDURE tWorm.paint_foot;
    BEGIN
      GOTOXY( ( pFoot^.pNext )^.x, ( pFoot^.pNext )^.y );
      WRITE( ASCII_SPACE );

    END;


{ ---------------------------------------------------------------------------------------------------------- INIT ----- }
BEGIN END.